/*
   Copyright 2004 Silicon Integrated Systems Corporation
   Revision:	1.00.00 <2004/12/20>

   Modified from the driver which is originally written by Donald Becker.

   This software may be used and distributed according to the terms
   of the GNU General Public License (GPL), incorporated in by reference.
   Drivers based on this skeleton fall under the GPL and must retain
   the authorship (implicit copyright) notice.

   References:
   SiS 965 MuTIOL Media I/O Programming Guide
   preliminary Rev. 1.0 Nov. 01, 2004
   http://www.sis.com.tw/support/databook.htm

   Rev 1.05    <2007/06/01>
   Rev 1.04    <2007/01/04>
   Rev 1.01    <2006/07/04>
   Rev 1.00.00 <2004/12/20> Chung-Hsien Chang <changch@sis.com> Initial release
*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/init.h>
#include <linux/mii.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/delay.h>
#include <linux/ethtool.h>
#include <linux/crc32.h>
#include <asm/processor.h>      /* Processor type for cache alignment. */
#include "sis190.h"

#define SIS190_MODULE_NAME "sis190"
#define SIS190_DRV_VERSION "v1.05.00 06/01/2007"
#define pci_module_init pci_register_driver

#define TRACE()	printk(KERN_INFO"[%s]: at line-%d\n", __FUNCTION__, __LINE__);
#define TRACE1(value) printk(KERN_INFO"[%s]: at line-%d, value = %x\n", __FUNCTION__, __LINE__, value);
#ifdef DEBUG
	#define DPRINTK(fmt, args...) printk(KERN_ERR "%s: " fmt, __FUNCTION__, ## args)
#else
	#define DPRINTK(fmt, args...)
#endif


static char version[] __devinitdata =
KERN_INFO "sis190.c: " SIS190_DRV_VERSION "\n";

static int max_interrupt_work = 40;
static int multicast_filter_limit = 128;

#define sis190_debug debug
static int sis190_debug = 0;

enum {
	SIS_190 = 0,
	SIS_191
};

static char * card_names[] = {
	"SiS 190 PCI Fast Ethernet Adapter",
	"SiS 191 PCI Gigabit Ethernet Adapter"
};

static struct pci_device_id sis190_pci_tbl [] = {
	{0x1039, 0x0190, PCI_ANY_ID, PCI_ANY_ID, 0, 0, SIS_190},
	{0x1039, 0x0191, PCI_ANY_ID, PCI_ANY_ID, 0, 0, SIS_191},
	{0,}
};
MODULE_DEVICE_TABLE (pci, sis190_pci_tbl);

static void sis190_read_mode(struct net_device *net_dev, int *speed, int *duplex);
static void handle_wired_interrupt(struct net_device *net_dev);


MODULE_AUTHOR("Silicon Integrated Systems");
MODULE_DESCRIPTION("SiS191/SiS190 1000/100/10 Ethernet driver");
MODULE_LICENSE("GPL");

MODULE_PARM_DESC(multicast_filter_limit, "SiS 191/190 maximum number of filtered multicast addresses");
MODULE_PARM_DESC(max_interrupt_work, "SiS 191/190 maximum events handled per interrupt");
MODULE_PARM_DESC(debug, "SiS 191/190 debug level (2-4)");


/**
 *	sis190_probe - Probe for sis190 device
 *	@pci_dev: the sis190 pci device
 *	@pci_id: the pci device ID
 *
 *	Check and probe sis190 net device for @pci_dev.
 *	Get mac address from EEPROM or APC,
 *	and assign SiS190-specific entries in the device structure.
 *	ie: sis190_open(), sis190_start_xmit(), sis190_close(), etc.
 */

static int __devinit sis190_probe (struct pci_dev *pci_dev, const struct pci_device_id *pci_id)
{
	struct sis190_private *sis_priv;
	struct net_device *net_dev;
	dma_addr_t ring_dma;
	void *ring_space;
	long ioaddr;
	int i, ret;
	u8 From;
	struct pci_dev *sis_sb = NULL;
	char *card_name = card_names[pci_id->driver_data];

/* when built into the kernel, we only print version if device is found */
#ifndef MODULE
	static int printed_version;
	if (!printed_version++)
		printk(version);
#endif

	/* setup various bits in PCI command register */
	ret = pci_enable_device(pci_dev);
	if(ret) return ret;

	i = pci_set_dma_mask(pci_dev, SIS190_DMA_MASK);
	if(i){
		printk(KERN_ERR "sis190.c: architecture does not support"
			"32bit PCI busmaster DMA\n");
		return i;
	}

	pci_set_master(pci_dev);

	net_dev = alloc_etherdev(sizeof(struct sis190_private));
	if (!net_dev)
		return -ENOMEM;
	SET_MODULE_OWNER(net_dev);
	SET_NETDEV_DEV(net_dev, &pci_dev->dev);

	/* We do a request_region() to register /proc/ioports info. */
	ioaddr = pci_resource_start(pci_dev, 1);
	ret = pci_request_regions(pci_dev, "sis190");
	if (ret)
		goto err_out;

	sis_priv = net_dev->priv;
	net_dev->base_addr = ioaddr;
	net_dev->irq = pci_dev->irq;
	sis_priv->pci_dev = pci_dev;
	sis_priv->vendor = pci_id->vendor;
	sis_priv->device = pci_id->device;
	sis_priv->wired_cnt = 0;
	spin_lock_init(&sis_priv->lock);

	// Find SiS South Bridge
	sis_sb = pci_find_slot (0, PCI_DEVFN(2, 0));
	if (sis_sb)
	{
		sis_priv->SB_id = sis_sb->device;
	}
	//lcc
	if (sis_priv->SB_id == 0x966)
	{
		pci_write_config_byte(pci_dev, 0xff, 0x20);
	}

	pci_set_drvdata(pci_dev, net_dev);

	// Allocate Tx Descriptors
	ring_space = pci_alloc_consistent(pci_dev, TX_TOTAL_SIZE + 15, &ring_dma);
	if (!ring_space) {
		ret = -ENOMEM;
		goto err_out_cleardev;
	}
	sis_priv->org_tx_ring = (HwTcb *)ring_space;
	sis_priv->org_tx_ring_dma = ring_dma;

	// Alignment for Tx descriptor ring
	sis_priv->tx_ring_dma = ((ring_dma + 0xf) & 0xfffffff0);
	sis_priv->tx_ring = (HwTcb *)(ring_space + (sis_priv->tx_ring_dma - ring_dma));

#ifdef DEBUG
	printk("sis_priv->org_tx_ring = %x, sis_priv->tx_ring = %x\n", (u32)sis_priv->org_tx_ring, (u32)sis_priv->tx_ring);
	printk("sis_priv->org_tx_ring_dma = %x, sis_priv->tx_ring_dma = %x\n",
			(u32)sis_priv->org_tx_ring_dma, (u32)sis_priv->tx_ring_dma);
#endif

	// Allocate Rx Descriptors
	ring_space = pci_alloc_consistent(pci_dev, RX_TOTAL_SIZE + 15, &ring_dma);
	if (!ring_space) {
		ret = -ENOMEM;
		goto err_unmap_tx;
	}
	sis_priv->org_rx_ring = (HwRfd *)ring_space;
	sis_priv->org_rx_ring_dma = ring_dma;

	// Alignment for Rx descriptor ring
	sis_priv->rx_ring_dma = ((ring_dma + 0xf) & 0xfffffff0);
	sis_priv->rx_ring = (HwRfd *)(ring_space + (sis_priv->rx_ring_dma - ring_dma));

#ifdef DEBUG
	printk("sis_priv->org_rx_ring = %x, sis_priv->rx_ring = %x\n", (u32)sis_priv->org_rx_ring, (u32)sis_priv->rx_ring);
	printk("sis_priv->org_rx_ring_dma = %x, sis_priv->rx_ring_dma = %x\n",
			(u32)sis_priv->org_rx_ring_dma, (u32)sis_priv->rx_ring_dma);
#endif

	/* The SiS190-specific entries in the device structure. */
	net_dev->open 			= &sis190_open;
	net_dev->hard_start_xmit	= &sis190_start_xmit;
	net_dev->stop 			= &sis190_close;
	net_dev->get_stats 		= &sis190_get_stats;
	net_dev->set_multicast_list 	= &set_rx_mode;
	net_dev->do_ioctl 		= &mii_ioctl;
	net_dev->tx_timeout 		= sis190_tx_timeout;
	net_dev->watchdog_timeo 	= TX_TIMEOUT;
	net_dev->ethtool_ops 		= &sis190_ethtool_ops;

	ret = register_netdev(net_dev);
	if (ret)
		goto err_unmap_rx;

	/* Get Mac address according to the from bit */
	pci_read_config_byte(pci_dev, 0x73, &From);
	ret = 0;

	if (From & BIT_0)
		ret = sis190_get_mac_addr_from_apc(pci_dev, net_dev);
	else
		ret = sis190_get_mac_addr_from_eeprom(pci_dev, net_dev);

	if (ret == 0) {
		ret = -ENODEV;
		goto err_out_unregister;
	}

	/* probe for mii transceiver */
	if (sis190_mii_probe(net_dev) == 0) {
		ret = -ENODEV;
		goto err_out_unregister;
	}

	/* print some information about our NIC */
	printk(KERN_INFO "%s: %s at %#lx, IRQ %d, ", net_dev->name,
	       card_name, ioaddr, net_dev->irq);

	for(i = 0; i < 5; i++)
		printk("%2.2x:", (u8)net_dev->dev_addr[i]);
	printk("%2.2x.\n", (u8)net_dev->dev_addr[i]);

	if(sis_priv->RGMII)
		printk(KERN_INFO "%s: Run in RGMII mode.\n", net_dev->name);
	else
		printk(KERN_INFO "%s: Run in MII/GMII mode.\n", net_dev->name);
	return 0;

 err_out_unregister:
 	unregister_netdev(net_dev);
 err_unmap_rx:
	pci_free_consistent(pci_dev, RX_TOTAL_SIZE + 15, sis_priv->org_rx_ring,
		sis_priv->org_rx_ring_dma);
 err_unmap_tx:
	pci_free_consistent(pci_dev, TX_TOTAL_SIZE + 15, sis_priv->org_tx_ring,
		sis_priv->org_tx_ring_dma);
 err_out_cleardev:
 	pci_set_drvdata(pci_dev, NULL);
	pci_release_regions(pci_dev);
 err_out:
	free_netdev(net_dev);
	return ret;
}

#if 0
static void SelectPhy(struct net_device * net_dev)
{
	struct sis190_private * sis_priv = net_dev->priv;
	u16  MdiControlReg = 0;
	u16  MdiStatusReg = 0;
	u16  MdiAdverReg = 0;
	u16  status;
	int  i = 0;

	// Read the MDI control register
	MdiControlReg = mdio_read(net_dev, sis_priv->cur_phy, MII_CONTROL);

	// Set/Clear bit unisolate this phy
	MdiControlReg &= ~MII_CNTL_ISOLATE;

	// issue the command to unisolate this Phy
	mdio_write(net_dev, sis_priv->cur_phy, MII_CONTROL, MdiControlReg);

	// sticky bits on link
	MdiStatusReg = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);
	MdiStatusReg = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);

	// Read the MDI control register
	MdiControlReg = mdio_read(net_dev, sis_priv->cur_phy, MII_CONTROL);

	// set Restart auto-negotiation
	MdiControlReg |= MII_CNTL_AUTO;
	MdiControlReg |= MII_CNTL_RST_AUTO;

	MdiAdverReg = mdio_read(net_dev, sis_priv->cur_phy, MII_ANADV);
	MdiAdverReg |= MII_NWAY_PAUSE;
	mdio_write(net_dev, sis_priv->cur_phy, MII_ANADV, MdiAdverReg);

	// restart the auto-negotion process
	mdio_write(net_dev, sis_priv->cur_phy, MII_CONTROL, MdiControlReg);
	udelay(200);

	for (i = 0; i < 350; i++) {
		status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);
		status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);
		if (status & MII_STAT_AUTO_DONE)
			break;
		msleep(10);
	}
}
#endif

static int Select_n_SetupPHY(struct net_device * net_dev)
{
	struct sis190_private * sis_priv = net_dev->priv;
	int speed, duplex;
	u16 status = 0, usValue;
	int complete = 0;
	int i;

	/* select default PHY for mac */
	sis_priv->mii = NULL;
	sis190_default_phy(net_dev);
#if 0
	// Attansic PHY F1
	if(sis_priv->mii->sid == PHY_ATHEROS_1 || sis_priv->mii->sid == PHY_ATHEROS_2 || sis_priv->mii->sid == PHY_ATHEROS_3) {
		mdio_write(net_dev, sis_priv->cur_phy, 0x14, 0x002e);
		mdio_write(net_dev, sis_priv->cur_phy, 0x1d, 0x0);
		udelay(200);
		mdio_write(net_dev, sis_priv->cur_phy, 0x1e, 0x034e);
		mdio_write(net_dev, sis_priv->cur_phy, MII_CONTROL, 0x9000);
	}
#endif
	/************************0719***************************/
	usValue = 0x01e1;		// auto mode
	mdio_write(net_dev, sis_priv->cur_phy, MII_ANADV, usValue);

	if (sis_priv->device == 0x191)
		mdio_write(net_dev, sis_priv->cur_phy, MII_ANGADV, 0x0200);
	else
		mdio_write(net_dev, sis_priv->cur_phy, MII_ANGADV, 0x0);

	if(( sis_priv->mii->sid == PHY_MRVL_1) || ( sis_priv->mii->sid == PHY_VITESSE)) {
		printk(KERN_INFO "====> [ %s ] Entering ResetPhy !!\n", __FUNCTION__);
	}
	status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);

#if 1
	for (i = 0; i < 350; i++) {
		status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);
		status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);
		if (status & MII_STAT_AUTO_DONE) {
			break;
		}
		msleep(1);
	}
#endif
	sis_priv->mii->status = status;

	if(status & MII_STAT_LINK) {
		complete = 1;
		sis_priv->state = 1;
	}
	else
		sis_priv->state = 0;

	if(complete) {
		sis190_read_mode(net_dev, &speed, &duplex);
		sis_priv->speed = speed;
		sis_priv->duplex = duplex;
		sis190_MacSpeed_Setup(net_dev, speed, duplex);
	}
	return 1;
}

/**
 *	sis190_mii_probe - Probe MII PHY for sis190
 *	@net_dev: the net device to probe for
 *
 *	Search for total of 32 possible mii phy addresses.
 *	Identify and set current phy if found one,
 *	return error if it failed to found.
 */

static int __init sis190_mii_probe (struct net_device * net_dev)
{
	struct sis190_private * sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	u16 mii_status;
	int phy_addr;
	u32 value32;


	sis_priv->mii = NULL;

      // Find a Valid PHY and Get PHY ID.
      // Workaround for Agere PHY ET1101B.
	value32 = inl(ioaddr+ StationCtrl);
	outl((value32 |0x0f000000), ioaddr +  StationCtrl);

	/* search for total of 32 possible mii phy addresses */
	for (phy_addr = 0; phy_addr < 32; phy_addr++)
	{
		struct mii_phy * mii_phy = NULL;
		int i;

		mii_phy = NULL;
		for(i = 0; i < 2; i++)
			mii_status = mdio_read(net_dev, phy_addr, MII_STATUS);

		if (mii_status == 0xffff || mii_status == 0x0000)
			/* the mii is not accessible, try next one */
			continue;

		if ((mii_phy = kmalloc(sizeof(struct mii_phy), GFP_KERNEL)) == NULL)
		{
			printk(KERN_INFO "Cannot allocate mem for struct mii_phy\n");
			/* Free previous find mii from first_mii */
			mii_phy = sis_priv->first_mii;
			while (mii_phy)
			{
				struct mii_phy *phy;
				phy = mii_phy;
				mii_phy = mii_phy->next;
				kfree(phy);
			}
			return 0;
		}

		mii_phy->phy_id0 = mdio_read(net_dev, phy_addr, MII_PHY_ID0);
		mii_phy->phy_id1 = mdio_read(net_dev, phy_addr, MII_PHY_ID1);
		mii_phy->phy_addr = phy_addr;
		mii_phy->status = mii_status;
		mii_phy->next = sis_priv->mii;
		sis_priv->mii = mii_phy;
		sis_priv->first_mii = mii_phy;

		for (i = 0; mii_chip_table[i].phy_id1; i++)
		{
			if ((mii_phy->phy_id0 == mii_chip_table[i].phy_id0 ) &&
			    ((mii_phy->phy_id1 & 0xFFF0) == mii_chip_table[i].phy_id1))
			{
				mii_phy->phy_types = mii_chip_table[i].phy_types;
				mii_phy->sid = mii_chip_table[i].sid;
				if (mii_chip_table[i].phy_types == MIX)
					mii_phy->phy_types =
						(mii_status & (MII_STAT_CAN_TX_FDX | MII_STAT_CAN_TX)) ? LAN : HOME;

				printk(KERN_INFO "%s: %s transceiver found at address %d.\n",
				       net_dev->name, mii_chip_table[i].name, phy_addr);
				break;
			}
		}

		if( !mii_chip_table[i].phy_id1 ) {
			printk(KERN_INFO "%s: Unknown PHY transceiver found at address %d.\n",
			       net_dev->name, phy_addr);
			mii_phy->phy_types = UNKNOWN;
		}
	}

	if (sis_priv->mii == NULL) {
		printk(KERN_INFO "%s: No MII transceivers found!\n", net_dev->name);
		return 0;
	}

	// Attansic PHY F1
	if(sis_priv->mii->sid == PHY_ATHEROS_1 || sis_priv->mii->sid == PHY_ATHEROS_2 || sis_priv->mii->sid == PHY_ATHEROS_3) {
		mdio_write(net_dev, sis_priv->cur_phy, 0x14, 0x002e);
		mdio_write(net_dev, sis_priv->cur_phy, 0x1d, 0x0);
		udelay(200);
		mdio_write(net_dev, sis_priv->cur_phy, 0x1e, 0x034e);
		mdio_write(net_dev, sis_priv->cur_phy, MII_CONTROL, 0x9000);
	}

	mii_status = mdio_read(net_dev, phy_addr, MII_STATUS);
	if ( !(sis_priv->mii->sid == PHY_AGERE) ||  (mii_status  & MII_STAT_LINK)) {
		outl(value32, ioaddr + StationCtrl);
	}
	return Select_n_SetupPHY(net_dev);
}

/**
 *	sis190_default_phy - Select default PHY for sis190 mac.
 *	@net_dev: the net device to probe for
 *
 *	Select first detected PHY with link as default.
 *	If no one is link on, select PHY whose types is HOME as default.
 *	If HOME doesn't exist, select LAN.
 */

static u16 sis190_default_phy(struct net_device * net_dev)
{
	struct sis190_private * sis_priv = net_dev->priv;
 	struct mii_phy *phy = NULL, *phy_home = NULL, *default_phy = NULL, *phy_lan = NULL;
	u16 status = 0;

	for( phy=sis_priv->first_mii; phy; phy=phy->next )
	{
		status = mdio_read(net_dev, phy->phy_addr, MII_STATUS);
		/* Link ON & Not select default PHY & not ghost PHY */
		 if ( (status & MII_STAT_LINK) && !default_phy && (phy->phy_types != UNKNOWN) ) {
		 	default_phy = phy;
		 	default_phy->status = status;	// log the status
		 }
		 else {
			if( phy->phy_types == HOME )
				phy_home = phy;
			else if (phy->phy_types == LAN)
				phy_lan = phy;
		 }
	}

	if( !default_phy && phy_home )
		default_phy = phy_home;

	else if( !default_phy && phy_lan )
		default_phy = phy_lan;

	else if ( !default_phy )
		default_phy = sis_priv->first_mii;

	if( sis_priv->mii != default_phy ) {
		sis_priv->mii = default_phy;
		sis_priv->cur_phy = default_phy->phy_addr;
	}

	status = mdio_read(net_dev, sis_priv->cur_phy, MII_CONTROL);
	status &= (~MII_CNTL_ISOLATE);

	mdio_write(net_dev, sis_priv->cur_phy, MII_CONTROL, status);
	status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);
	status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);

	return status;
}



#if 0
/**
 *	sis190_reset_phy - reset sis190 mii phy.
 *	@net_dev: the net device to write
 *	@phy_addr: default phy address
 *
 *	Some specific phy can't work properly without reset.
 *	This function will be called during initialization.
 */

static void sis190_reset_phy(struct net_device *net_dev, BOOL bAutoNeg, BOOL bWaitAutoNeg)
{
	struct sis190_private *sis_priv = net_dev->priv;
	int i = 0;
	u16 status;

	status = mdio_read(net_dev, sis_priv->cur_phy, MII_CONTROL);
	TRACE();

	if(bAutoNeg) {
		// Reset the Phy, enable auto-negotiation, and restart auto-negotiation. 0x9200
		status |= (MII_CNTL_RESET | MII_CNTL_AUTO | MII_CNTL_RST_AUTO);
	}
	else {
		status |= MII_CNTL_RESET;
	}

	Marvell_Patch(net_dev);

	// Write the MDI control register with our new Phy configuration
	mdio_write( net_dev, sis_priv->cur_phy, MII_CONTROL, status);

	Vitesse_Patch(net_dev);
	TruePhyConfig_Patch(net_dev);

	if (bAutoNeg && bWaitAutoNeg) {
		// wait for auto-negotiation to complete (up to 3.5 seconds)
		for (i = 35; i > 0; i--) {
			status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);
			status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);
			if (status & MII_STAT_AUTO_DONE)
				break;
			udelay(100);
		}
	}
}
#endif

/**
 *	sis190_open - open sis190 device
 *	@net_dev: the net device to open
 *
 *	Do some initialization and start net interface.
 *	enable interrupts and set sis190 timer.
 */

static int
sis190_open(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	int speed, duplex;
	int ret;

	/* Soft reset the chip */
	sis190_reset(net_dev);

	if (!sis_priv->autong_complete)
	{
		TRACE();
		sis190_read_mode(net_dev, &speed, &duplex);
		sis_priv->speed = speed;
		sis_priv->duplex = duplex;
		sis190_MacSpeed_Setup(net_dev, speed, duplex);
	}
	else
	{
		sis190_MacSpeed_Setup(net_dev, sis_priv->speed, sis_priv->duplex);
	}

	ret = request_irq(net_dev->irq, &sis190_interrupt, SA_SHIRQ, net_dev->name, net_dev);

	if (ret)
		return ret;

	outl(0, ioaddr + RxMacAddr);
	sis190_init_rxfilter(net_dev);
	sis190_init_tx_ring(net_dev);
	sis190_init_rx_ring(net_dev);
	set_rx_mode(net_dev);

	/* Enable all known interrupts by setting the interrupt mask. */
	outl(INT_MASK, ioaddr + IntMask);

	// Enable Tx and Rx
	outl((inl(ioaddr + TxCtrl) | BIT_0), ioaddr + TxCtrl);
	outl((inl(ioaddr + RxCtrl) | BIT_0 | BIT_4 | BIT_6), ioaddr + RxCtrl);

	netif_start_queue(net_dev);

	return 0;
}

/**
 *	sis190_init_rxfilter - Initialize the Rx filter
 *	@net_dev: the net device to initialize for
 *
 *	Set receive filter address to our MAC address
 *	and enable packet filtering.
 */

static void
sis190_init_rxfilter (struct net_device * net_dev)
{
	long ioaddr = net_dev->base_addr;
	u16 rfcrSave;
	u32 i;
	rfcrSave = inw(ioaddr + RxMacCtrl);

	/* disable packet filtering before setting filter */
	outl((rfcrSave & ~BIT_8_11), ioaddr + RxMacCtrl);

	/* load MAC addr to filter data register */
	for (i = 0 ; i < 6 ; i++) {
		u8 w;
		w = (u8) *((u8 *)(net_dev->dev_addr)+i);
		outb(w, ioaddr + RxMacAddr + i);
	}

	/* enable packet filtering */
	outw(rfcrSave, ioaddr + RxMacCtrl);
}

/**
 *	sis190_init_tx_ring - Initialize the Tx descriptor ring
 *	@net_dev: the net device to initialize for
 *
 *	Initialize the Tx descriptor ring,
 */

static void
sis190_init_tx_ring(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	int i;

	sis_priv->tx_full = 0;
	sis_priv->dirty_tx = sis_priv->cur_tx = 0;

	for (i = 0; i < NUM_TX_DESC; i++) {
		sis_priv->tx_skbuff[i] = NULL;

		sis_priv->tx_ring[i].PktSize = 0;
		sis_priv->tx_ring[i].cmdsts = 0;
		sis_priv->tx_ring[i].bufptr = 0;

		if(i == (NUM_TX_DESC - 1)) {
			sis_priv->tx_ring[i].EOD = BIT_31;
		}
		else {
			sis_priv->tx_ring[i].EOD = 0;
		}
	}

	/* load Transmit Descriptor Register */
	outl(sis_priv->tx_ring_dma, ioaddr + TxBase);
	if (sis190_debug > 2)
		printk(KERN_INFO "%s: TX descriptor register loaded with: %8.8x\n",
		       net_dev->name, inl(ioaddr + TxBase));
}

static void
sis190_reset_rx_ring(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	int i;

	sis_priv->cur_rx = 0;
	sis_priv->dirty_rx = 0;

	/* init RX descriptor */
	for (i = 0; i < NUM_RX_DESC; i++) {
		sis_priv->rx_ring[i].StsSize = 0;
		sis_priv->rx_ring[i].PktInfo = 0xc0000000;
		if(i == (NUM_RX_DESC - 1)) {
			sis_priv->rx_ring[i].EOD = BIT_31;
		}
		else {
			sis_priv->rx_ring[i].EOD = 0;
		}
		sis_priv->rx_ring[i].EOD |= (RX_BUF_SIZE & BIT_3_15);
	}

	sis_priv->dirty_rx = (unsigned int) (i - NUM_RX_DESC);

	/* load Receive Descriptor Register */
	outl(sis_priv->rx_ring_dma, ioaddr + RxBase);
	if (sis190_debug > 2)
		printk(KERN_INFO "%s: RX descriptor register loaded with: %8.8x\n",
		       net_dev->name, inl(ioaddr + RxBase));

}

/**
 *	sis190_init_rx_ring - Initialize the Rx descriptor ring
 *	@net_dev: the net device to initialize for
 *
 *	Initialize the Rx descriptor ring,
 *	and pre-allocate recevie buffers (socket buffer)
 */

static void
sis190_init_rx_ring(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	int i;

	sis_priv->cur_rx = 0;
	sis_priv->dirty_rx = 0;

	/* init RX descriptor */
	for (i = 0; i < NUM_RX_DESC; i++) {
		sis_priv->rx_skbuff[i] = NULL;

		sis_priv->rx_ring[i].StsSize = 0;
		sis_priv->rx_ring[i].PktInfo = 0xc0000000;
		sis_priv->rx_ring[i].bufptr = 0;

		if(i == (NUM_RX_DESC - 1)) {
			sis_priv->rx_ring[i].EOD = BIT_31;
		}
		else {
			sis_priv->rx_ring[i].EOD = 0;
		}
	}

	/* allocate sock buffers */
	for (i = 0; i < NUM_RX_DESC; i++) {
		struct sk_buff *skb;

		if ((skb = dev_alloc_skb(RX_BUF_SIZE)) == NULL) {
			/* not enough memory for skbuff, this makes a "hole"
			   on the buffer ring, it is not clear how the
			   hardware will react to this kind of degenerated
			   buffer */
			printk(KERN_INFO "%s: Not enough RX skb buffer\n", net_dev->name);
			break;
		}
		skb->dev = net_dev;
		sis_priv->rx_skbuff[i] = skb;
		sis_priv->rx_ring[i].EOD |= (RX_BUF_SIZE & BIT_3_15);
		sis_priv->rx_ring[i].bufptr = pci_map_single(sis_priv->pci_dev,
                        skb->tail, RX_BUF_SIZE, PCI_DMA_FROMDEVICE);
	}
	sis_priv->dirty_rx = (unsigned int) (i - NUM_RX_DESC);

	/* load Receive Descriptor Register */
	outl(sis_priv->rx_ring_dma, ioaddr + RxBase);
	if (sis190_debug > 2)
		printk(KERN_INFO "%s: RX descriptor register loaded with: %8.8x\n",
		       net_dev->name, inl(ioaddr + RxBase));
}

/**
 *	sis190_MacSpeed_Setup - Set the media mode of mac register.
 *	@ioaddr: the address of the device
 *	@speed : the transmit speed to be determined
 *	@duplex: the duplex mode to be determined
 *
 *	Set the media mode of mac register StationCtrl according to
 *	speed and duplex of phy.
 */

static void sis190_MacSpeed_Setup (struct net_device *net_dev, int speed, int duplex)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	u32 reg, value32 = 0;

	// Configure MAC Speed
	reg = inl(ioaddr + StationCtrl);
	reg &= ~(BIT_10_12 | BIT_24_25);

	if (sis_priv->mii->status & MII_STAT_LINK) {
		switch(speed) {
			case HW_SPEED_1000_MBPS:
				reg |= (BIT_10_11 | BIT_24_25);  // Tx select internal clock for 1000Mbps
				value32 = 0x02220000;
				break;

			case HW_SPEED_100_MBPS:
				reg |= BIT_11;
				value32 = 0x01110000;
				break;

			case HW_SPEED_10_MBPS:
				value32 = 0x01110000;
				reg |= BIT_10;
				break;
		}
	}
	// setup Interrupt Control
	outl(value32 | 0xf0000000, ioaddr + IntCtrl);

	// Duplex mode
	if (duplex == FDX_CAPABLE_FULL_SELECTED) {
		reg |= BIT_12;
		value32 = inl(ioaddr + RxMacCtrl);
		value32 &= ~BIT_3;
		outl(value32, ioaddr + RxMacCtrl);
	}
	else {
		value32 = inl(ioaddr + RxMacCtrl);
		value32 |= BIT_3;
		outl(value32, ioaddr + RxMacCtrl);
	}

	if(sis_priv->RGMII) {
		// RGMII Select Tx internal clock
		reg |= BIT_24_25;
	}
	else if ((sis_priv->SB_id == 0x965) && (speed == HW_SPEED_1000_MBPS)) {
		// External Clock
            reg &= ~BIT_24_25;
            reg |= BIT_25;
	}
	// write the StationControl Register
	outl(reg, ioaddr + StationCtrl);


	if(sis_priv->RGMII)
	{
		if (sis_priv->SB_id == 0x965) {
			outl(0x0441, ioaddr + TxMacDelayCtrl);
			outl(0x0440, ioaddr + TxMacDelayCtrl);
		}
		else {
			outl(0x04A1, ioaddr + TxMacDelayCtrl);
			outl(0x04A0, ioaddr + TxMacDelayCtrl);
		}

		switch(sis_priv->mii->sid) {
			case PHY_ATHEROS_1:	// For Attansic F1 PHY Need longer Tx delay
				if (sis_priv->SB_id < 0x968)
					outl(0x04C9, ioaddr + TxMacDelayCtrl);
				break;

			case PHY_BCM_1:	// For Broadcom BCM5461 PHY, set Tx Delay in RGMII mode
				mdio_write(net_dev, sis_priv->cur_phy, 0x18, 0xf1c7);
				udelay(400);
				mdio_write(net_dev, sis_priv->cur_phy, 0x1c, 0x8c00);
				break;

			case PHY_BCM_2:
				mdio_write(net_dev, sis_priv->cur_phy, 0x18, 0x81e7);
				udelay(400);
				mdio_write(net_dev, sis_priv->cur_phy, 0x1c, 0x8804);
				break;
		}
	}


	value32 = (BIT_8_9 | BIT_13);

	if (duplex == FDX_CAPABLE_FULL_SELECTED) {
		outl(0x0, ioaddr + TxLimit);
	}
	else {
		value32 |= (BIT_10_12 | BIT_2);
		outl(0xf, ioaddr + TxLimit);
		value32 &= ~BIT_0;
		if (speed == HW_SPEED_1000_MBPS)
			value32 |= BIT_14_15;
	}
	outl( inl(ioaddr + TxMacCtrl) | value32, ioaddr + TxMacCtrl);
}

/**
 *	sis190_read_mode - read media mode for sis190 internal phy
 *	@net_dev: the net device to read mode for
 *	@speed  : the transmit speed to be determined
 *	@duplex : the duplex mode to be determined
 *
 *	The capability of remote end will be put in mii register autorec
 *	after auto-negotiation. Use AND operation to get the upper bound
 *	of speed and duplex between two ends.
 */

static void sis190_read_mode(struct net_device *net_dev, int *speed, int *duplex)
{
	struct sis190_private *sis_priv = net_dev->priv;
	int phy_addr = sis_priv->cur_phy;
	u32 status;
	u16 autoadv, autorec, autoexp, gigadv, gigrec;
	int i = 0;

	while (i++ < 2)
		status = mdio_read(net_dev, phy_addr, MII_STATUS);

	/* log phy status to MII */
	sis_priv->mii->status = status;

	*speed = HW_SPEED_100_MBPS;
	*duplex = FDX_CAPABLE_FULL_SELECTED;

	if (!(status & MII_STAT_LINK)) {
		return;
	}


	/* AutoNegotiate completed */
	autorec = mdio_read(net_dev, phy_addr, MII_ANLPAR);	// Link partner register
	autoadv = mdio_read(net_dev, phy_addr, MII_ANADV);
	autoexp = mdio_read(net_dev, phy_addr, MII_ANEXT);

	// Link partner has 1000 Mbps capability
	if((sis_priv->device == 0x191) && (autorec & MII_NWAY_NP))
	{
		gigadv = mdio_read(net_dev, phy_addr, MII_ANGADV);
		gigrec = mdio_read(net_dev, phy_addr, MII_ANGLPAR);
		status = (gigadv & (gigrec >> 2));
		if(status & BIT_9)
		{
			*speed = HW_SPEED_1000_MBPS;
			*duplex = FDX_CAPABLE_FULL_SELECTED;
			DPRINTK("G-Full\n");
		}
		else if (status & BIT_8)
		{
			*speed = HW_SPEED_1000_MBPS;
			*duplex = FDX_CAPABLE_HALF_SELECTED;
			DPRINTK("G-Half\n");
		}
	}
	else
	{
		status = autoadv & autorec;
		if (status & (MII_NWAY_TX | MII_NWAY_TX_FDX))
			*speed = HW_SPEED_100_MBPS;
		if (status & ( MII_NWAY_TX_FDX | MII_NWAY_T_FDX))
			*duplex = FDX_CAPABLE_FULL_SELECTED;
		DPRINTK("100\n");
	}

	sis_priv->autong_complete = 1;

	printk(KERN_INFO "%s: Media Link On %d Mbps %s-duplex \n",
	       net_dev->name,
	       *speed,
	       *duplex == FDX_CAPABLE_FULL_SELECTED ?
	       "full" : "half");
	return;
}

/**
 *	sis190_tx_timeout - sis190 transmit timeout routine
 *	@net_dev: the net device to transmit
 *
 *	print transmit timeout status
 *	disable interrupts and do some tasks
 */

static void sis190_tx_timeout(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	unsigned long flags;
	int i;
	u32 reg;

	printk(KERN_INFO "%s: Transmit timeout, status %8.8x %8.8x \n",
	       net_dev->name, inl(ioaddr + TxCtrl), inl(ioaddr + TxSts));

	/* use spinlock to prevent interrupt handler accessing buffer ring */
	spin_lock_irqsave(&sis_priv->lock, flags);

	/* Disable interrupts by clearing the interrupt mask. */
	outl(0x0000, ioaddr + IntMask);

	// Disable Tx
	reg = (inl(ioaddr + TxCtrl) & ~(BIT_0 | BIT_4));
	outl(reg, ioaddr + TxCtrl);
	udelay(10);

	if(inl(ioaddr + TxCtrl) & BIT_0)
		printk(KERN_INFO "%s: Can not disable Tx, status %8.8x %8.8x \n",
		       net_dev->name, inl(ioaddr + TxCtrl), inl(ioaddr + TxSts));

	/* discard unsent packets */
	sis_priv->dirty_tx = sis_priv->cur_tx = 0;
	for (i = 0; i < NUM_TX_DESC; i++) {
		struct sk_buff *skb = sis_priv->tx_skbuff[i];

		if (skb) {
			pci_unmap_single(sis_priv->pci_dev,
				sis_priv->tx_ring[i].bufptr, skb->len,
				PCI_DMA_TODEVICE);
			dev_kfree_skb_irq(skb);
			sis_priv->tx_skbuff[i] = NULL;
			sis_priv->stats.tx_dropped++;
		}

		sis_priv->tx_ring[i].PktSize = 0;
		sis_priv->tx_ring[i].cmdsts = 0;
		sis_priv->tx_ring[i].bufptr = 0;
		sis_priv->tx_ring[i].bufptr &= 0x80000000;
	}
	sis_priv->tx_full = 0;
	netif_wake_queue(net_dev);

	net_dev->trans_start = jiffies;

	/* Enable Tx */
	outl((reg | BIT_0), ioaddr + TxCtrl);

	/* Enable all known interrupts by setting the interrupt mask. */
	outl(INT_MASK, ioaddr + IntMask);

	spin_unlock_irqrestore(&sis_priv->lock, flags);

	return;
}

/**
 *	sis190_start_xmit - sis190 start transmit routine
 *	@skb: socket buffer pointer to put the data being transmitted
 *	@net_dev: the net device to transmit with
 *
 *	Set the transmit buffer descriptor,
 *	and write TxStart to enable transmit state machine.
 *	tell upper layer if the buffer is full
 */

static int
sis190_start_xmit(struct sk_buff *skb, struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	unsigned int  entry;
	unsigned long flags;

	/* Don't transmit data before the complete of auto-negotiation */
	if(!sis_priv->autong_complete){
		netif_stop_queue(net_dev);
		return 1;
	}

	spin_lock_irqsave(&sis_priv->lock, flags);

	/* Calculate the next Tx descriptor entry. */
	entry = sis_priv->cur_tx % NUM_TX_DESC;
	sis_priv->tx_skbuff[entry] = skb;

#ifdef DEBUG
	printk("sis_priv->cur_tx = %x, sis_priv->dirty_tx = %x\n", sis_priv->cur_tx, sis_priv->dirty_tx);
	printk("entry = %2.2x, TxDescAddr = %8.8x\n", entry, inl(ioaddr + TxSts));
#endif

	/* set the transmit buffer descriptor and enable Transmit State Machine */
	sis_priv->tx_ring[entry].bufptr = pci_map_single(sis_priv->pci_dev,
		skb->data, skb->len, PCI_DMA_TODEVICE);
	sis_priv->tx_ring[entry].PktSize = (skb->len & 0xffff);
	sis_priv->tx_ring[entry].EOD |= (skb->len & 0xffff);
	sis_priv->tx_ring[entry].cmdsts = (PADEN | CRCEN | DEFEN | THOL3 | TXINT);
	if(sis_priv->duplex == FDX_CAPABLE_HALF_SELECTED)
	{
		sis_priv->tx_ring[entry].cmdsts |= (COLEN | CRSEN | BKFEN);
		if(sis_priv->speed == HW_SPEED_1000_MBPS)
			sis_priv->tx_ring[entry].cmdsts |= (EXTEN | BSTEN);
	}
	sis_priv->tx_ring[entry].cmdsts |= TXOWN;

	outl(BIT_4 | BIT_6 | inl(ioaddr + TxCtrl), ioaddr + TxCtrl);

	sis_priv->cur_tx ++;

	if((sis_priv->cur_tx - sis_priv->dirty_tx) >= (NUM_TX_DESC - 4))
	{
		/* dirty_tx is met in the cycle of cur_tx, buffer full */
		printk(KERN_INFO "%s: Tx buffers is full, queue stoped !!\n", net_dev->name);
		sis_priv->tx_full = 1;
		netif_stop_queue(net_dev);
	}

	spin_unlock_irqrestore(&sis_priv->lock, flags);

	net_dev->trans_start = jiffies;

	if (sis190_debug > 3)
		printk(KERN_INFO "%s: Queued Tx packet at %p size %d "
		       "to slot %d.\n",
	       net_dev->name, skb->data, (int)skb->len, entry);

	return 0;
}

static void handle_wired_interrupt(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	u32 reg;
	int i;
	u32 ulvalue;

	netif_carrier_off(net_dev);
	netif_stop_queue(net_dev);

	/* Disable interrupts by clearing the interrupt mask. */
	outl(0x0000, ioaddr + IntMask);
	reg = (inl(ioaddr + TxCtrl) & ~(BIT_0 | BIT_4));
	outl(reg, ioaddr + TxCtrl);
	udelay(100);

	if(inl(ioaddr + TxCtrl) & BIT_0)
		printk(KERN_INFO "%s: Can not disable Tx, status %8.8x %8.8x \n",
	net_dev->name, inl(ioaddr + TxCtrl), inl(ioaddr + TxSts));

	/* discard unsent packets */
	sis_priv->dirty_tx = sis_priv->cur_tx = 0;
	for (i = 0; i < NUM_TX_DESC; i++) {
		struct sk_buff *skb = sis_priv->tx_skbuff[i];

		if (skb) {
			pci_unmap_single(sis_priv->pci_dev,
				sis_priv->tx_ring[i].bufptr, skb->len,
				PCI_DMA_TODEVICE);
			dev_kfree_skb_irq(skb);
			sis_priv->tx_skbuff[i] = NULL;
			sis_priv->stats.tx_dropped++;
		}

		sis_priv->tx_ring[i].PktSize = 0;
		sis_priv->tx_ring[i].cmdsts = 0;
		sis_priv->tx_ring[i].bufptr = 0;
		sis_priv->tx_ring[i].bufptr &= 0x80000000;
	}
	sis_priv->tx_full = 0;
	net_dev->trans_start = jiffies;

	/* Stop the chip's Tx and Rx Status Machine */
	ulvalue = inl(ioaddr + IntCtrl) | BIT_15;
	outl(ulvalue, ioaddr + IntCtrl);
	udelay(50);
	ulvalue &= ~BIT_15;
	outl(ulvalue, ioaddr + IntCtrl);

	sis190_init_rxfilter(net_dev);
	sis190_init_tx_ring(net_dev);
	sis190_reset_rx_ring(net_dev);
	set_rx_mode(net_dev);

	/* Enable all known interrupts by setting the interrupt mask. */
	outl(INT_MASK, ioaddr + IntMask);

	// Enable Tx and Rx
	outl((inl(ioaddr + TxCtrl) | BIT_0), ioaddr + TxCtrl);
	outl((inl(ioaddr + RxCtrl) | BIT_0 | BIT_4 | BIT_6), ioaddr + RxCtrl);

	netif_carrier_on(net_dev);
	netif_start_queue(net_dev);
}

static void handle_linkchange_interrupt(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	u32 reg;
	int i;
	u16 status;
	u32 ulvalue;

	status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);
	status = mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS);

//	if (status & MII_STAT_LINK){
	if (sis_priv->state == 1) {
//		printk (KERN_INFO"%s: Link OFF -> ON\n", net_dev->name);
		Select_n_SetupPHY(net_dev);

		outl(0, ioaddr + RxMacAddr);
		sis190_init_rxfilter(net_dev);
		sis190_init_tx_ring(net_dev);
		sis190_reset_rx_ring(net_dev);
		set_rx_mode(net_dev);

		/* Enable all known interrupts by setting the interrupt mask. */
		outl(INT_MASK, ioaddr + IntMask);

		// Enable Tx and Rx
		outl((inl(ioaddr + TxCtrl) | BIT_0), ioaddr + TxCtrl);
		outl((inl(ioaddr + RxCtrl) | BIT_0 | BIT_4 | BIT_6), ioaddr + RxCtrl);

		netif_carrier_on(net_dev);
		netif_start_queue(net_dev);
	}
	else {
//		printk (KERN_INFO"%s: Link ON -> OFF\n", net_dev->name);

		netif_carrier_off(net_dev);
		netif_stop_queue(net_dev);

		/* Disable interrupts by clearing the interrupt mask. */
		outl(0x0000, ioaddr + IntMask);
		reg = (inl(ioaddr + TxCtrl) & ~(BIT_0 | BIT_4));
		outl(reg, ioaddr + TxCtrl);
		udelay(100);

		if(inl(ioaddr + TxCtrl) & BIT_0)
			printk(KERN_INFO "%s: Can not disable Tx, status %8.8x %8.8x \n",
		net_dev->name, inl(ioaddr + TxCtrl), inl(ioaddr + TxSts));

		/* discard unsent packets */
		sis_priv->dirty_tx = sis_priv->cur_tx = 0;
		for (i = 0; i < NUM_TX_DESC; i++) {
			struct sk_buff *skb = sis_priv->tx_skbuff[i];

			if (skb) {
				pci_unmap_single(sis_priv->pci_dev,
					sis_priv->tx_ring[i].bufptr, skb->len,
					PCI_DMA_TODEVICE);
				dev_kfree_skb_irq(skb);
				sis_priv->tx_skbuff[i] = NULL;
				sis_priv->stats.tx_dropped++;
			}

			sis_priv->tx_ring[i].PktSize = 0;
			sis_priv->tx_ring[i].cmdsts = 0;
			sis_priv->tx_ring[i].bufptr = 0;
			sis_priv->tx_ring[i].bufptr &= 0x80000000;
		}
		sis_priv->tx_full = 0;
		net_dev->trans_start = jiffies;

		/* Stop the chip's Tx and Rx Status Machine */
		ulvalue = inl(ioaddr + IntCtrl) | BIT_15;
		outl(ulvalue, ioaddr + IntCtrl);
		udelay(50);
		ulvalue &= ~BIT_15;
		outl(ulvalue, ioaddr + IntCtrl);
	}
}

/**
 *	sis190_interrupt - sis190 interrupt handler
 *	@irq: the irq number
 *	@dev_instance: the client data object
 *	@regs: snapshot of processor context
 *
 *	The interrupt handler does all of the Rx thread work,
 *	and cleans up after the Tx thread
 */
static irqreturn_t sis190_interrupt(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct net_device *net_dev = dev_instance;
	struct sis190_private *sis_priv = net_dev->priv;
	int boguscnt = max_interrupt_work;
	long ioaddr = net_dev->base_addr;
	u32 status, mask;
	unsigned int handled = 0;

	spin_lock (&sis_priv->lock);

	// Disable interrupts
	outl(0, ioaddr + IntMask);
	do {
		status = inl(ioaddr + IntSrc);
		mask = INT_MASK;

		if ((status & mask) == 0) {
			/* nothing intresting happened */
			break;
		}
		handled = 1;
		outl(status, ioaddr + IntSrc);

		if (status & BIT_16) {
			if ( mdio_read(net_dev, sis_priv->cur_phy, MII_STATUS)
					& MII_STAT_LINK )
			{
				if (sis_priv->state == 0) {
//					printk("%s: Link changed interrupt, link_on, nic state = %d !!\n",
//						net_dev->name, sis_priv->state);
					sis_priv->state = 1;
					handle_linkchange_interrupt(net_dev);
				}
				else {
//					printk ("wired - link_on, state = 1 \n");
					sis_priv->wired_cnt++;
					if (sis_priv->wired_cnt == 2) {
						handle_wired_interrupt(net_dev);
						sis_priv->wired_cnt = 0;
					}
				}
			}
			else {
				if (sis_priv->state == 1) {
//					printk("%s: Link changed interrupt, link_off, nic state = %d !!\n",
//						net_dev->name, sis_priv->state);
					sis_priv->state = 0;
					handle_linkchange_interrupt(net_dev);
				}
//				else
//					printk ("wired - link_off, state = 0 \n");
			}
		}

		if (status & BIT_0)
			printk(KERN_INFO "%s: Tx Halt interrupt issued !!\n", net_dev->name);

		if (status & BIT_1)
			printk(KERN_INFO "%s: Rx Halt interrupt issued !!\n", net_dev->name);

		if (status & (BIT_2 | BIT_3))
			/* Tx interrupt */
			sis190_finish_xmit(net_dev);

		if (status & (BIT_6 | BIT_7))
			/* Rx interrupt */
			sis190_rx(net_dev);

		if (--boguscnt < 0) {
			printk(KERN_INFO "%s: Too much work at interrupt, "
			       "interrupt status = %#8.8x.\n",
			       net_dev->name, status);
			break;
		}
	} while (1);

	// Enable interrupts
	outl(INT_MASK, ioaddr + IntMask);
	spin_unlock (&sis_priv->lock);
	return IRQ_RETVAL(handled);
}

/**
 *	sis190_rx - sis190 receive routine
 *	@net_dev: the net device which receives data
 *
 *	Process receive interrupt events,
 *	put buffer to higher layer and refill buffer pool
 *	Note: This fucntion is called by interrupt handler,
 *	don't do "too much" work here
 */

static int sis190_rx(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	unsigned int entry = sis_priv->cur_rx % NUM_RX_DESC;
	u32 descount;
	u32 rx_status = sis_priv->rx_ring[entry].StsSize;
	u32 rx_info = sis_priv->rx_ring[entry].PktInfo;

	if (sis190_debug > 5)
		printk(KERN_INFO "sis190_rx, cur_rx:%4.4d, dirty_rx:%4.4d "
		       "status:0x%8.8x\n",
		       sis_priv->cur_rx, sis_priv->dirty_rx, rx_status);

	while (!(rx_info & RXOWN)) {
		unsigned int rx_size;

		rx_size = (rx_status & 0xffff);
		descount = ((rx_status & BIT_24_30) >> 24);
		if(descount > 1)
			printk(KERN_INFO "%s: Rx Desc Count %d in a packet.\n", net_dev->name, descount);

		if ((rx_status & (COLON|NIBON|OVRUN|MIIER|LIMIT|SHORT|ABORT)) ||
		    !(rx_status & CRCOK))
		{
			/* corrupted packet received */
			if (sis190_debug > 3)
				printk(KERN_INFO "%s: Corrupted packet "
				       "received, buffer status = 0x%8.8x.\n",
				       net_dev->name, rx_status);
			sis_priv->stats.rx_errors++;
			if (rx_status & OVRUN)
				sis_priv->stats.rx_over_errors++;
			if (rx_status & (SHORT|LIMIT))
				sis_priv->stats.rx_length_errors++;
			if (rx_status & (MIIER | NIBON | COLON))
				sis_priv->stats.rx_frame_errors++;
			if (!(rx_status & CRCOK))
				sis_priv->stats.rx_crc_errors++;
			/* reset buffer descriptor state */
			sis_priv->rx_ring[entry].StsSize = 0;
			sis_priv->rx_ring[entry].PktInfo = 0xc0000000;
		} else {
			struct sk_buff * skb;

			/* This situation should never happen, but due to
			   some unknow bugs, it is possible that
			   we are working on NULL sk_buff :-( */
			if (sis_priv->rx_skbuff[entry] == NULL) {
				printk(KERN_INFO "%s: NULL pointer "
				       "encountered in Rx ring, skipping\n",
				       net_dev->name);
				break;
			}

			pci_unmap_single(sis_priv->pci_dev,
				sis_priv->rx_ring[entry].bufptr, RX_BUF_SIZE,
				PCI_DMA_FROMDEVICE);
			/* give the socket buffer to upper layers */
			skb = sis_priv->rx_skbuff[entry];
			skb_put(skb, rx_size);
			skb->protocol = eth_type_trans(skb, net_dev);
			netif_rx(skb);

			/* some network statistics */
			if ((rx_status & BCAST) == MCAST)
				sis_priv->stats.multicast++;
			net_dev->last_rx = jiffies;
			sis_priv->stats.rx_bytes += rx_size;
			sis_priv->stats.rx_packets++;

			/* refill the Rx buffer, what if there is not enought memory for
			   new socket buffer ?? */
			if ((skb = dev_alloc_skb(RX_BUF_SIZE)) == NULL) {
				/* not enough memory for skbuff, this makes a "hole"
				   on the buffer ring, it is not clear how the
				   hardware will react to this kind of degenerated
				   buffer */
				printk(KERN_INFO "%s: Memory squeeze,"
				       "deferring packet.\n",
				       net_dev->name);
				sis_priv->rx_skbuff[entry] = NULL;
				/* reset buffer descriptor state */
				sis_priv->rx_ring[entry].StsSize = 0;
				sis_priv->rx_ring[entry].PktInfo = 0;
				sis_priv->rx_ring[entry].bufptr = 0;
				sis_priv->stats.rx_dropped++;
				break;
			}
			skb->dev = net_dev;
			sis_priv->rx_skbuff[entry] = skb;
			sis_priv->rx_ring[entry].StsSize = 0;
			sis_priv->rx_ring[entry].PktInfo = 0xc0000000;
                	sis_priv->rx_ring[entry].bufptr =
				pci_map_single(sis_priv->pci_dev, skb->tail,
					RX_BUF_SIZE, PCI_DMA_FROMDEVICE);
			sis_priv->dirty_rx++;
		}
		sis_priv->cur_rx++;
		entry = sis_priv->cur_rx % NUM_RX_DESC;
		rx_status = sis_priv->rx_ring[entry].StsSize;
		rx_info = sis_priv->rx_ring[entry].PktInfo;
	} // while

	/* refill the Rx buffer, what if the rate of refilling is slower than
	   consuming ?? */
	for (;sis_priv->cur_rx - sis_priv->dirty_rx > 0; sis_priv->dirty_rx++) {
		struct sk_buff *skb;

		entry = sis_priv->dirty_rx % NUM_RX_DESC;

		if (sis_priv->rx_skbuff[entry] == NULL) {
			if ((skb = dev_alloc_skb(RX_BUF_SIZE)) == NULL) {
				/* not enough memory for skbuff, this makes a "hole"
				   on the buffer ring, it is not clear how the
				   hardware will react to this kind of degenerated
				   buffer */
				printk(KERN_INFO "%s: Memory squeeze,"
				       "deferring packet.\n",
				       net_dev->name);
				sis_priv->stats.rx_dropped++;
				break;
			}
			skb->dev = net_dev;
			sis_priv->rx_skbuff[entry] = skb;
                	sis_priv->rx_ring[entry].bufptr =
				pci_map_single(sis_priv->pci_dev, skb->tail,
					RX_BUF_SIZE, PCI_DMA_FROMDEVICE);
		}
	}
	/* re-enable the potentially idle receive state matchine */
	outl(BIT_4 | BIT_6 | inl(ioaddr + RxCtrl), ioaddr + RxCtrl );

	return 0;
}

/**
 *	sis190_finish_xmit - finish up transmission of packets
 *	@net_dev: the net device to be transmitted on
 *
 *	Check for error condition and free socket buffer etc
 *	schedule for more transmission as needed
 *	Note: This fucntion is called by interrupt handler,
 *	don't do "too much" work here
 */

static void sis190_finish_xmit (struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;

	for (; sis_priv->dirty_tx != sis_priv->cur_tx; sis_priv->dirty_tx++) {
		struct sk_buff *skb;
		unsigned int entry;
		u32 tx_status;
		u32 pkt_size;

		entry = sis_priv->dirty_tx % NUM_TX_DESC;
		tx_status = sis_priv->tx_ring[entry].cmdsts;
		pkt_size = (sis_priv->tx_ring[entry].EOD & 0xffff);

		if (tx_status & TXOWN) {
			/* The packet is not transmitted yet (owned by hardware) !
			   Note: the interrupt is generated only when Tx Machine
			   is idle, so this is an almost impossible case */
			break;
		}

		if (tx_status & BIT_16_19) {
			/* packet unsuccessfully transmitted */
			if (sis190_debug > 3)
				printk(KERN_INFO "%s: Transmit "
				       "error, Tx status %8.8x.\n",
				       net_dev->name, tx_status);
			sis_priv->stats.tx_errors++;
			if (tx_status & BIT_17)
				sis_priv->stats.tx_fifo_errors++;
			if (tx_status & BIT_18)
				sis_priv->stats.tx_aborted_errors++;
			if (tx_status & BIT_16)
				sis_priv->stats.tx_carrier_errors++;
			if (tx_status & BIT_19)
				sis_priv->stats.tx_window_errors++;
		} else {
			/* packet successfully transmitted */
			sis_priv->stats.collisions += ((tx_status & 0xffff) - 1);
			sis_priv->stats.tx_bytes += pkt_size;
			sis_priv->stats.tx_packets++;
		}
		/* Free the original skb. */
		skb = sis_priv->tx_skbuff[entry];
		pci_unmap_single(sis_priv->pci_dev,
			sis_priv->tx_ring[entry].bufptr, skb->len,
			PCI_DMA_TODEVICE);
		dev_kfree_skb_irq(skb);
		sis_priv->tx_skbuff[entry] = NULL;
		sis_priv->tx_ring[entry].bufptr = 0;
		sis_priv->tx_ring[entry].cmdsts = 0;
		sis_priv->tx_ring[entry].PktSize = 0;
		sis_priv->tx_ring[entry].EOD &= 0x80000000;
	}

	if (sis_priv->tx_full && netif_queue_stopped(net_dev) &&
	    sis_priv->cur_tx - sis_priv->dirty_tx < NUM_TX_DESC - 4) {
		/* The ring is no longer full, clear tx_full and schedule more transmission
		   by netif_wake_queue(net_dev) */
		printk(KERN_INFO "%s: Some Tx buffers are released, queue restarted !!\n", net_dev->name);
		sis_priv->tx_full = 0;
		netif_wake_queue (net_dev);
	}
}

/**
 *	sis190_close - close sis190 device
 *	@net_dev: the net device to be closed
 *
 *	Disable interrupts, stop the Tx and Rx Status Machine
 *	free Tx and RX socket buffer
 */

static int
sis190_close(struct net_device *net_dev)
{
	long ioaddr = net_dev->base_addr;
	struct sis190_private *sis_priv = net_dev->priv;
	struct sk_buff *skb;
	int i;
	u32 ulvalue;

	netif_stop_queue(net_dev);

	/* Disable interrupts by clearing the interrupt mask. */
	outl(0x0000, ioaddr + IntMask);
	udelay(2000);

	/* Stop the chip's Tx and Rx Status Machine */
	ulvalue = inl(ioaddr + IntCtrl) | BIT_15;
	outl(ulvalue, ioaddr + IntCtrl);
	udelay(50);
	ulvalue &= ~BIT_15;
	outl(ulvalue, ioaddr + IntCtrl);

	free_irq(net_dev->irq, net_dev);

	/* Free Tx and RX skbuff */
	for (i = 0; i < NUM_RX_DESC; i++) {
		skb = sis_priv->rx_skbuff[i];
		if (skb) {
			pci_unmap_single(sis_priv->pci_dev,
				sis_priv->rx_ring[i].bufptr,
				RX_BUF_SIZE, PCI_DMA_FROMDEVICE);
			dev_kfree_skb(skb);
			sis_priv->rx_skbuff[i] = NULL;
		}
	}
	for (i = 0; i < NUM_TX_DESC; i++) {
		skb = sis_priv->tx_skbuff[i];
		if (skb) {
			pci_unmap_single(sis_priv->pci_dev,
				sis_priv->tx_ring[i].bufptr, skb->len,
				PCI_DMA_TODEVICE);
			dev_kfree_skb(skb);
			sis_priv->tx_skbuff[i] = NULL;
		}
	}

	/* Green! Put the chip in low-power mode. */

	return 0;
}

/**
 *	sis190_get_drvinfo - Return information about driver
 *	@net_dev: the net device to probe
 *	@info: container for info returned
 *
 *	Process ethtool command such as "ehtool -i" to show information
 */

static void sis190_get_drvinfo(struct net_device *net_dev,
			       struct ethtool_drvinfo *info)
{
 	struct sis190_private *sis_priv = net_dev->priv;

	strcpy (info->driver, SIS190_MODULE_NAME);
	strcpy (info->version, SIS190_DRV_VERSION);
	strcpy (info->bus_info, pci_name(sis_priv->pci_dev));
}

static struct ethtool_ops sis190_ethtool_ops = {
	.get_drvinfo =		sis190_get_drvinfo,
};

/**
 *	mii_ioctl - process MII i/o control command
 *	@net_dev: the net device to command for
 *	@rq: parameter for command
 *	@cmd: the i/o command
 *
 *	Process MII command like read/write MII register
 */

static int mii_ioctl(struct net_device *net_dev, struct ifreq *rq, int cmd)
{
	struct sis190_private *sis_priv = net_dev->priv;
	struct mii_ioctl_data *data = if_mii(rq);

	switch(cmd) {
		case SIOCGMIIPHY:		/* Get address of MII PHY in use. */
		#ifdef DEBUG
			printk("Get Address of MII PHY\n");
		#endif
			data->phy_id = sis_priv->mii->phy_addr;
			return 0;

		case SIOCGMIIREG:		/* Read MII PHY register. */
			data->val_out = mdio_read(net_dev, (data->phy_id & 0x1f), (data->reg_num & 0x1f));
		#ifdef DEBUG
			printk("Read PHY register, Phy = %d, Reg = %d, Data = %x\n", data->phy_id, data->reg_num, data->val_out);
		#endif
			return 0;

		case SIOCSMIIREG:		/* Write MII PHY register. */
			if (!capable(CAP_NET_ADMIN))
				return -EPERM;
		#ifdef DEBUG
			printk("Write PHY register, Phy = %d, Reg = %d, Value = %x\n", data->phy_id, data->reg_num, data->val_in);
		#endif
			mdio_write(net_dev, data->phy_id & 0x1f, data->reg_num & 0x1f, data->val_in);
			return 0;
		default:
			return -EOPNOTSUPP;
	}
}

/**
 *	sis190_get_stats - Get sis190 read/write statistics
 *	@net_dev: the net device to get statistics for
 *
 *	get tx/rx statistics for sis190
 */

static struct net_device_stats *
sis190_get_stats(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	return &sis_priv->stats;
}

/**
 *	sis190_mcast_bitnr - compute hashtable index
 *	@addr: multicast address
 *
 *	SiS 190 uses the last sigificant 6 bits to index a 64 bits multicast
 *	hash table.
 */

static inline u16 sis190_mcast_bitnr(u8 *addr)
{

	u32 crc = ether_crc(6, addr);

	return ((u16)(crc & BIT_0_5));
}

/**
 *	set_rx_mode - Set SiS190 receive mode
 *	@net_dev: the net device to be set
 *
 *	Set SiS190 receive mode for promiscuous, multicast, or broadcast mode.
 *	And set the appropriate multicast filter.
 *	Multicast hash table has the size of 64 bits for SiS190.
 */

static void set_rx_mode(struct net_device *net_dev)
{
	long ioaddr = net_dev->base_addr;
	u32 mc_filter[2] = {0};	/* 64 bits multicast hash table */
	int i;
	u16 rx_mode;

	outw((inw(ioaddr + RxMacCtrl) & ~BIT_8_11), ioaddr + RxMacCtrl);

	if (net_dev->flags & IFF_PROMISC) {
		/* Accept any kinds of packets */
		rx_mode = BIT_8_11;
		for (i = 0; i < 2; i++)
			mc_filter[i] = 0xffffffff;
	}
	else if ((net_dev->mc_count > multicast_filter_limit) || (net_dev->flags & IFF_ALLMULTI)) {

		/* too many multicast addresses or accept all multicast packet */
		rx_mode = BIT_11 | BIT_10 | BIT_9;
		for (i = 0; i < 2; i++)
			mc_filter[i] = 0xffffffff;
	} else {

		/* Accept Broadcast packet, destination address matchs our MAC address,
		   use Receive Filter to reject unwanted MCAST packet */
		struct dev_mc_list *mclist;
		rx_mode = BIT_11 | BIT_10 | BIT_9;

		for (i = 0, mclist = net_dev->mc_list; mclist && i < net_dev->mc_count;
		     i++, mclist = mclist->next)
		{
			u16 bit_nr =	sis190_mcast_bitnr(mclist->dmi_addr);
			if(bit_nr < 0x20)
				mc_filter[0] |= (1 << bit_nr);
			else
				mc_filter[1] |= (1 << (bit_nr - 0x20));
		}
	}

	/* update Multicast Hash Table in Receive Filter */
	outl(mc_filter[0], ioaddr + RxHashTbl);
	outl(mc_filter[1], ioaddr + RxHashTbl + 4);

	outw( (inw(ioaddr + RxMacCtrl) | rx_mode), ioaddr + RxMacCtrl);

	/* sis190 is capatable of looping back packet at MAC level for debugging purpose */
	if (net_dev->flags & IFF_LOOPBACK) {
		u32 txcr_saved, rxcr_saved;
		/* We must disable Tx/Rx before setting loopback mode */
		txcr_saved = inl(ioaddr + TxCtrl);
		outl((txcr_saved & ~BIT_0), ioaddr + TxMacCtrl);
		rxcr_saved = inl(ioaddr + RxCtrl);
		outl((rxcr_saved & ~BIT_0), ioaddr + RxMacCtrl);
		/* enable loopback */
		outl(inl(ioaddr + StationCtrl) | BIT_31, ioaddr + StationCtrl);
		/* restore Tx/Rx Ctrl registers */
		outl(txcr_saved, ioaddr + TxMacCtrl);
		outl(rxcr_saved, ioaddr + RxMacCtrl);
	}

	return;
}

/**
 *	sis190_reset - Reset sis190 MAC
 *	@net_dev: the net device to reset
 *
 *	Reset sis190 MAC and wait until finished
 *	Initialize all MAC registers.
 */

static void sis190_reset(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;

	// Reset Tx & Rx
	outl(0x00001c00, ioaddr + TxCtrl);
	if (sis_priv->SB_id == 0x968)
		outl(0x081e1c00, ioaddr + RxCtrl);
	else
		outl(0x001e1c00, ioaddr + RxCtrl);

	outl(BIT_15, ioaddr + IntCtrl);
	udelay(100);
	outl(0, ioaddr + IntCtrl);

	// Disable Interrupt
	outl(0, ioaddr + IntMask);
	outl(0xffffffff, ioaddr + IntSrc);

	// Give initial value for all MAC registers
	outl(0, ioaddr + TxBase);
	outl(0, ioaddr + rsv0);
	outl(0, ioaddr + RxBase);
	outl(0, ioaddr + rsv1);
	outl(0xffc00000, ioaddr + PowCtrl);
	outl(0, ioaddr + rsv2);
	if(sis_priv->RGMII)
		outl(0x04008001, ioaddr + StationCtrl);
	else
		outl(0x04000001, ioaddr + StationCtrl);

	if (sis_priv->SB_id >= 0x966 )
		outl(0x80000000, ioaddr + GIoCR);
	else
		outl(0, ioaddr + GIoCR);
	outl(0, ioaddr + GIoCtrl);
	outl(0x00002364, ioaddr + TxMacCtrl);
	outl(0x0000000f, ioaddr + TxLimit);
	outl(0, ioaddr + TxMacDelayCtrl);
	outl(0, ioaddr + rsv3);

	if (sis_priv->SB_id == 0x968)
		outw(0x22d2, ioaddr + RxMacCtrl);
	else
		outw(0x0252, ioaddr + RxMacCtrl);
	outl(0, ioaddr + RxHashTbl);
	outl(0, ioaddr + RxHashTbl + 4);
	outl(0x80ff0000, ioaddr + RxWolCtrl);
	outl(0x80ff0000, ioaddr + RxWolData);
	outl(0x0, ioaddr + RxMpsCtrl);
	if (sis_priv->SB_id == 0x968)
		outl(0xcf, ioaddr + rsv4);
	else
		outl(0x0, ioaddr + rsv4);
}

/**
 *	sis190_remove - Remove sis190 device
 *	@pci_dev: the pci device to be removed
 *
 *	remove and release SiS190 net device
 */

static void __devexit sis190_remove(struct pci_dev *pci_dev)
{
	struct net_device *net_dev = pci_get_drvdata(pci_dev);
	struct sis190_private * sis_priv = net_dev->priv;
	struct mii_phy *phy = NULL;

	while (sis_priv->first_mii) {
		phy = sis_priv->first_mii;
		sis_priv->first_mii = phy->next;
		kfree(phy);
	}

	pci_free_consistent(pci_dev, RX_TOTAL_SIZE + 15, sis_priv->org_rx_ring,
		sis_priv->org_rx_ring_dma);
	pci_free_consistent(pci_dev, TX_TOTAL_SIZE + 15, sis_priv->org_tx_ring,
		sis_priv->org_tx_ring_dma);

	unregister_netdev(net_dev);
	free_netdev(net_dev);
	pci_release_regions(pci_dev);
	pci_set_drvdata(pci_dev, NULL);
}

static struct pci_driver sis190_pci_driver = {
	.name		= SIS190_MODULE_NAME,
	.id_table	= sis190_pci_tbl,
	.probe		= sis190_probe,
	.remove	= __devexit_p(sis190_remove),
};

static int __init sis190_init_module(void)
{
/* when a module, this is printed whether or not devices are found in probe */
#ifdef MODULE
	printk(version);
#endif

	return pci_module_init(&sis190_pci_driver);
}

static void __exit sis190_cleanup_module(void)
{
#ifdef DEBUG
	printk("SiS19x ==> sis19x_cleanup_module !!\n");
#endif

	pci_unregister_driver(&sis190_pci_driver);
}

module_init(sis190_init_module);
module_exit(sis190_cleanup_module);

